-- Proofs about booleans
-- Author :: Ulf Norell
-- Created:: 2004-12-14

-- # import Property
-- # import Bool

-- import Eq
-- import Ord

-- Equality ---------------------------------------------------------------
propIf :
    (A : Set) -> (P : A -> Prop) ->
    (x : Bool) -> (y, z : A) ->
    (Holds x        -> P y) ->
    (Holds (not x)  -> P z) ->
    P (if x y z)
propIf A P true  y z ptrue pfalse = ptrue tt
propIf A P false y z ptrue pfalse = pfalse tt

-- Holds x <=> x == true --------------------------------------------------
propHoldsTrue : (x:Bool) -> Holds x <=> x == true
propHoldsTrue true  = fol-plugin()
propHoldsTrue false = fol-plugin(Bool)

propHoldsTrueE : (x:Bool) -> Holds x ==> x == true
propHoldsTrueE x = andE1 (propHoldsTrue x)

propHoldsTrueI : (x:Bool) -> Holds x <== x == true
propHoldsTrueI x = andE2 (propHoldsTrue x)

-- Not (Holds x) <=> x == false -------------------------------------------
propHoldsFalse : (x:Bool) -> Not (Holds x) <=> x == false
propHoldsFalse true  = fol-plugin(Bool)
propHoldsFalse false = fol-plugin()

propHoldsFalseE : (x:Bool) -> Not (Holds x) ==> x == false
propHoldsFalseE x = andE1 (propHoldsFalse x)

propHoldsFalseI : (x:Bool) -> Not (Holds x) <== x == false
propHoldsFalseI x = andE2 (propHoldsFalse x)

propHoldsNot : (x : Bool) -> Holds (not x) <=> Not (Holds x)
propHoldsNot true = fol-plugin()
propHoldsNot false = fol-plugin()

propHoldsNotE : (x : Bool) -> Holds (not x) ==> Not (Holds x)
propHoldsNotE x = andE1 (propHoldsNot x)

propHoldsNotI : (x : Bool) -> Holds (not x) <== Not (Holds x)
propHoldsNotI x = andE2 (propHoldsNot x)

-- (&&) <=> (/\) ----------------------------------------------------------
propHoldsAnd : (x, y : Bool) -> Holds (x && y) <=> Holds x /\ Holds y
propHoldsAnd true  y = fol-plugin() -- andI (impI (andI tt)) (impI andE2)
propHoldsAnd false y = fol-plugin()

propHoldsAndE : (x, y : Bool) -> Holds (x && y) ==> Holds x /\ Holds y
propHoldsAndE x y = andE1 (propHoldsAnd x y)

propHoldsAndI : (x, y : Bool) -> Holds (x && y) <== Holds x /\ Holds y
propHoldsAndI x y = andE2 (propHoldsAnd x y)

-- (||) <=> (\/) ----------------------------------------------------------
propHoldsOr : (x, y : Bool) -> Holds (x || y) <=> Holds x \/ Holds y
propHoldsOr true = fol-plugin()
propHoldsOr false = fol-plugin()

propHoldsOrE : (x, y : Bool) -> Holds (x || y) ==> Holds x \/ Holds y
propHoldsOrE x y = andE1 (propHoldsOr x y)

propHoldsOrI : (x, y : Bool) -> Holds (x || y) <== Holds x \/ Holds y
propHoldsOrI x y = andE2 (propHoldsOr x y)

-- Associativity of (&&) and (||) -----------------------------------------
propAndAssoc : (x, y, z : Bool) -> (x && y) && z == x && (y && z)
propAndAssoc true  y z  = fol-plugin()
propAndAssoc false y z  = fol-plugin()

propOrAssoc : (x, y, z : Bool) -> (x || y) || z == x || (y || z)
propOrAssoc true  y z   = fol-plugin()
propOrAssoc false y z   = fol-plugin()

-- instance Equ Bool where
eqBool : Bool -> Bool -> Bool
eqBool true  x  = x
eqBool false x  = not x

{-
propBoolEq : IsDecidableEquality eqBool
propBoolEq true x  = fol-plugin(propHoldsTrue)
propBoolEq false x = andI
   fol-plugin(propHoldsNotE, propHoldsFalseE)
   fol-plugin(propHoldsNotI, propHoldsFalseI)
  

equBool : Eq Bool
equBool = struct {eq=eqBool; proofs = fol-plugin(propBoolEq) }

-- instance Ord Bool where
leqBool : Bool -> Bool -> Bool
leqBool false x = true
leqBool true  x = x

eqBoolRel   = decidableRelation eqBool
leqBoolRel  = decidableRelation leqBool

propReflLeqBool : Reflexive' eqBoolRel leqBoolRel
propReflLeqBool false y = \_ -> tt
propReflLeqBool true  y = \p -> p

propAntiSymLeqBool : AntiSymmetric' eqBoolRel leqBoolRel
propAntiSymLeqBool false false  = \_ _ -> tt
propAntiSymLeqBool true false   = \p _ -> p
propAntiSymLeqBool false true   = \_ p -> p
propAntiSymLeqBool true true    = \_ _ -> tt

propTransLeqBool : Transitive leqBoolRel
propTransLeqBool false y z        = \_ _ -> tt
propTransLeqBool true false z     = fol-plugin()
propTransLeqBool true true false  = fol-plugin()
propTransLeqBool true true true   = \_ _ -> tt

propTotalLeqBool : Total leqBoolRel
propTotalLeqBool false y    = fol-plugin()
propTotalLeqBool true false = fol-plugin()
propTotalLeqBool true true  = fol-plugin()

ordBool : Ord Bool
ordBool = struct
  leq = leqBool
  equ = equBool
  proofs = struct
    refl    = propReflLeqBool
    antisym = propAntiSymLeqBool
    trans   = propTransLeqBool
    total   = propTotalLeqBool
-}
