-- Proofs about natural numbers
-- Author :: Ulf Norell
-- Created:: 2004-12-14

-- # import Property
-- # import Nat
-- # import Bool.Proofs
-- # import Monoid
-- # import Ord
-- # import Eq

-- Equality ---------------------------------------------------------------
infix 5 `EqNat`
EqNat = (==) {Nat}

-- Proofs about addition --------------------------------------------------

--- 0 + n == n
propPlusZero : (n : Nat) -> 0 + n `EqNat` n
propPlusZero 0        = reflexivity 0
propPlusZero (suc  n) = fol-plugin(propPlusZero n)

-- suc  m + n == suc  (m + n)
propPlusSuc  : (n, m : Nat) -> suc  n + m `EqNat` suc  (n + m)
propPlusSuc  n 0        = reflexivity (suc  n)
propPlusSuc  n (suc  m) = fol-plugin(propPlusSuc  n m)

-- associativity
propPlusAssoc : (x, y, z : Nat) -> (x + y) + z `EqNat` x + (y + z)
propPlusAssoc x y 0         = reflexivity (x + y)
propPlusAssoc x y (suc  z)  = fol-plugin(propPlusAssoc x y z)

-- commutativity
propPlusCommute : (n, m : Nat) -> n + m `EqNat` m + n
propPlusCommute 0        m  = propPlusZero m
propPlusCommute (suc  n) m  = fol-plugin(propPlusSuc  n m, propPlusCommute n m)

-- (Nat, 0, +) is a commutative monoid
monoidLawsZeroPlus : MonoidLaws Nat 0 (+)
monoidLawsZeroPlus =  struct
  id    = struct {left = propPlusZero; right = reflexivity }
  assoc = propPlusAssoc

monoidZeroPlus : Monoid Nat
monoidZeroPlus = struct
  ops    = struct {zero=0;plus=(+)}
  proofs = monoidLawsZeroPlus

commutativeMonoidLawsZeroPlus : CommutativeMonoidLaws Nat 0 (+)
commutativeMonoidLawsZeroPlus = struct
  monoidLaws = monoidLawsZeroPlus
  commute    = propPlusCommute

commutativeMonoidZeroPlus : CommutativeMonoid Nat
commutativeMonoidZeroPlus = struct
  ops    = struct {zero=0;plus=(+)}
  proofs = commutativeMonoidLawsZeroPlus

natEqRel  = decidableRelation eqNat
natLeqRel = decidableRelation (<=)

-- (==) is in fact Eq Nat
propNatEq : IsDecidableEquality eqNat
propNatEq 0        0        = fol-plugin()
propNatEq (suc  n) 0        = fol-plugin(Nat)
propNatEq 0        (suc  m) = fol-plugin(Nat)
propNatEq (suc  n) (suc  m) = fol-plugin(Nat, propNatEq n m)

-- Reflexive (<=)
propNatLeqRefl : Reflexive' natEqRel natLeqRel
propNatLeqRefl 0 0                = \_ -> tt
propNatLeqRefl (suc  n) 0         = \p -> p
propNatLeqRefl 0 (suc  m)         = \_ -> tt
propNatLeqRefl (suc  n) (suc  m)  = propNatLeqRefl n m

-- Antisym (<=)
propNatLeqAntiSym : AntiSymmetric' natEqRel natLeqRel
propNatLeqAntiSym 0 0 p q               = tt
propNatLeqAntiSym (suc  n) 0 p q        = p
propNatLeqAntiSym 0 (suc  m) p q        = q
propNatLeqAntiSym (suc  n) (suc  m) p q = propNatLeqAntiSym n m p q

-- Transitive (<=)
propNatLeqTrans : Transitive natLeqRel
propNatLeqTrans 0 y z p q                       = tt
propNatLeqTrans (suc  x) 0 z p q                = fol-plugin(p)
propNatLeqTrans (suc  x) (suc  y) 0 p q         = fol-plugin(q)
propNatLeqTrans (suc  x) (suc  y) (suc  z) p q  = propNatLeqTrans x y z p q

-- Total (<=)
propNatLeqTotal : Total natLeqRel
propNatLeqTotal 0 y               = fol-plugin()
propNatLeqTotal x 0               = fol-plugin()
propNatLeqTotal (suc  x) (suc  y) = propNatLeqTotal x y

equNat : Eq Nat
equNat = struct
  eq     = eqNat
  proofs = fol-plugin(propNatEq)

ordNat : Ord Nat
ordNat = struct
  leq = (<=)
  equ = equNat
  proofs = struct
    refl    = propNatLeqRefl
    antisym = propNatLeqAntiSym
    trans   = propNatLeqTrans
    total   = propNatLeqTotal

