-- Relations.

-- # import Property
-- # import Bool
-- # import Equality

Relation : Set -> Type
Relation A = A -> A -> Prop

Reflexive (A : Set) : Relation A -> Prop
Reflexive {A} (#) = (x : A) -> x # x

Reflexive' (A : Set) : Relation A -> Relation A -> Prop
Reflexive' {A} (~=) (#) = (x, y : A) -> x ~= y -> x # y

AntiReflexive (A : Set) : Relation A -> Prop
AntiReflexive {A} (#) = (x, y : A) -> Not (x # x)

AntiReflexive' (A : Set) : Relation A -> Relation A -> Prop
AntiReflexive' {A} (~=) (#) = (x, y : A) -> x ~= y -> Not (x # y)

Symmetric (A : Set) : Relation A -> Prop
Symmetric {A} (#) = (x, y : A) -> x # y -> y # x

AntiSymmetric (A : Set) : Relation A -> Prop
AntiSymmetric {A} (#) = (x, y : A) -> x # y -> y # x -> x == y

AntiSymmetric' (A : Set) : Relation A -> Relation A -> Prop
AntiSymmetric' {A} (~=) (#) = (x, y : A) -> x # y -> y # x -> x ~= y

Transitive (A : Set) : Relation A -> Prop
Transitive {A} (#) = (x, y, z : A) -> x # y -> y # z -> x # z

Total (A : Set) : Relation A -> Prop
Total {A} (#) = (x, y : A) -> x # y \/ y # x

EquivalenceRelation (A : Set) : Relation A -> Prop
EquivalenceRelation {A} R = sig
  refl   : Reflexive R
  sym    : Symmetric R
  trans  : Transitive R

PartialEquivalenceRelation (A : Set) : Relation A -> Type
PartialEquivalenceRelation R = sig
  sym    : Symmetric R
  trans  : Transitive R

TotalOrder (A : Set) : Relation A -> Relation A -> Type
TotalOrder E R = sig
  refl     : Reflexive' E R
  antisym  : AntiSymmetric' E R
  trans    : Transitive R
  total    : Total R

PartialOrder (A : Set) : Relation A -> Relation A -> Type
PartialOrder E R = sig
  refl     : Reflexive' E R
  antisym  : AntiSymmetric' E R
  trans    : Transitive R

-- The Prop Eq is an equivalence relation ---------------------------------

equivEq (A : Set) : EquivalenceRelation (==)
equivEq {A} = struct
    refl = refl
    sym  = sym
    trans= trans
  where
    refl : (x : A) -> x == x
    refl = reflexivity

    sym : Symmetric (==)
    sym = \x -> \y -> \pxy -> symmetry x y `impE` pxy

    trans : Transitive (==)
    trans = \x y z -> \pxy pyz -> transitivity x y z `impE` pxy `impE` pyz

-- Decidable relations ----------------------------------------------------

DecidableRelation : Set -> Set
DecidableRelation A = A -> A -> Bool

decidableRelation (A : Set) : DecidableRelation A -> Relation A
decidableRelation (#) = \x y -> Holds (x # y)

IsDecidableEquality (A : Set) : DecidableRelation A -> Prop
IsDecidableEquality {A} (#) = (x, y : A) -> Holds (x # y) <=> x == y
