\subsection{Problems in algebra}

 As a simple test, we represent the notion of boolean ring, as being a ring
(not necessarily commutative) in which any element is idempotent. This can be
represented as

\begin{code}
-- # import ExamplePrelude

isRing : (X:Set) -> (X->X->X) -> (X->X->X) ->
                      (X->X) -> X -> X -> Prop
isRing X (+) (*) minus Zero One =
 (x:X) -> (y:X) -> (z:X) ->
     (   x + y == y + x
      /\ x + Zero == x
      /\ x + (minus x) == Zero
      /\ x + (y+z) == (x+y)+z
      /\ x * (y+z) == (x*y) + (x*z)
      /\ (y+z)* x  == (y*x) + (z*x)
      /\ x * One == x
      /\ One * x == x
      /\ x * (y*z) == (x*y)*z
     )

isBool : (X:Set) -> (X->X->X) -> Prop
isBool X (*) = (x:X)-> (x*x) == x

R : Set
(+) : R -> R -> R
(*) : R -> R -> R
minus : R -> R 
Zero : R
One : R

axR1 : isRing R (+) (*) minus Zero One 
axR2 : isBool R (*)
\end{code}

%isBool : (X:Set) -> (X->X->X) -> (X->X->X) ->
%                      (X->X) -> X -> X -> Prop
%isBool X (+) (*) minus Zero One =
%  isRing X (+) (*) minus Zero One 
%  /\ ((x:X)-> Id X (x*x) x)

 Gandalf has no problem to show that such in any such ring, we have $x+x=0$, 
and also that the ring is commutative. We can then define a relation
$x\leq y$ by $xy = x$ and then ask to prove that $x=y \IFF (x\leq y\wedge y\leq x)$
holds. Gandalf needs more than the default time limit for this, but we can then
use the let construct to give simpler subproblems that are solved directly

\begin{code}
thm : (x:R) -> x + x == Zero
thm x = fol-plugin(axR1,axR2)

thm1 : (x:R) -> (y:R) -> x * y == y * x
thm1 x y = fol-plugin(axR1,axR2)

(<=) : R -> R -> Prop
x <= y = x * y == x

thm2 : (x:R) -> (y:R) ->  (x == y <=> (x <= y /\ y <= x))
thm2 x y = 
 let 
   rem1 : x == y ==> (x <= y /\ y <= x)
   rem1 = impI (\ h -> fol-plugin(axR1,axR2,h))
   rem2 : (x<=y /\ y<=x) ==> x == y
   rem2 = impI (\ h -> fol-plugin(axR1,axR2,h))
 in 
  fol-plugin(rem1,rem2)
\end{code}

%%% Local Variables:: 
%%% mode:: latex
%%% TeX-master:: "lfprop"
%%% End:: 
