
module Plugins.FOL.Errors where

import Text.PrettyPrint

import Syntax.Internal	( Value, Type, Term )
import Syntax.Pretty	( pretty )

import Utils.Error
import Plugins.FOL.Types
import Plugins.FOL.Constants

-- Whats and whens --------------------------------------------------------

whenTranslatingToFOL :: Term -> Doc
whenTranslatingToFOL e =
    fsep $ pwords "when translating" ++ [pretty e] ++
	   pwords "to first order logic"

higherOrderQuantification :: Value -> Doc
higherOrderQuantification t =
    fsep $ pwords "quantification over" ++ [pretty t] ++
	   pwords "is not first order"

-- Errors -----------------------------------------------------------------

notAFOLPredicateError :: Value -> ErrorFun
notAFOLPredicateError v e =
    e	{ errWhat   = fsep $ pwords "the type" ++ [pretty v] ++
			     pwords "is not a valid FOL predicate"
	, errWhy    = fsep $ pwords "it is not of the form (x::A) -> .. -> Holds p"
	}

notAFOLPropositionError :: Value -> ErrorFun
notAFOLPropositionError v e =
    e	{ errWhat   = fsep $ pwords "the proposition" ++ [pretty v] ++
			     pwords "is not first order"
	}

notAFOLTermError :: Value -> ErrorFun
notAFOLTermError v e =
    e	{ errWhat   = fsep $ pwords "the value" ++ [pretty v] ++
			     pwords "is not a first order term"
	}

noInferenceForFOLError :: ErrorFun
noInferenceForFOLError e =
    e	{ errWhat   = fsep $ pwords "type inference failed"
	, errWhy    = fsep $ pwords "the type of a first order goal cannot be inferred"
	}

failedToProveFirstOrderGoalsError :: [(FOLGoal, Maybe CounterModel)] -> ErrorFun
failedToProveFirstOrderGoalsError gs e =
    e	{ errWhen   = fsep $ pwords "solving first order goals"
	, errWhat   = vcat $ fsep (pwords "the prover failed to prove") : map display gs
	}
    where
	display (FOLGoal name _ _ _ _,Nothing) = nest 2 $ text $ "Goal " ++ name
	display (FOLGoal name _ _ _ _, Just m) = nest 2 $
				vcat
				[ text $ "Goal " ++ name
				, nest 2 $ text "Counter model:"
				, nest 4 $ vcat $ map text m
				]

notOfTypePropError :: Value -> Type -> ErrorFun
notOfTypePropError v t e =
    e	{ errWhat   = fsep $ pwords "the type" ++ [pretty v] ++
			     pwords "is not a first order formula"
	, errWhy    = fsep $ pwords "it has type" ++ [pretty t] ++
			     pwords "when it should have type " ++ pwords fol_Prop
	}

